@@ -1,11 +1,11 @@
 ;;; LaHaShem HaAretz U'Mloah
 
-;;; Stalin 0.8 - A global optimizing compiler for Scheme
+;;; Stalin 0.9 - A global optimizing compiler for Scheme
 ;;; Copyright 1993, 1994, and 1995 University of Toronto. All rights reserved.
 ;;; Copyright 1996 Technion. All rights reserved.
 ;;; Copyright 1996 and 1997 University of Vermont. All rights reserved.
-;;; Copyright 1997, 1998, and 1999 NEC Research Institute, Inc. All rights
-;;; reserved.
+;;; Copyright 1997, 1998, 1999, and 2000 NEC Research Institute, Inc. All
+;;; rights reserved.
 
 ;;; This program is free software; you can redistribute it and/or
 ;;; modify it under the terms of the GNU General Public License
@@ -34,15 +34,15 @@
 
 ;;; Stalin outputs its internal database of assertions about the program being
 ;;; compiled in a form that Emacs can read. Emacs commands allow one to
-;;; conveniently query this data base to assertain properties of the program.
+;;; conveniently query this database to assertain properties of the program.
 ;;; Stalin has its own READ procedure to keep track of the line and character
 ;;; positions of each S expression in the source file being compiled so that
-;;; it can generate more informative error messages. These Emacs command use
+;;; it can generate more informative error messages. These Emacs commands use
 ;;; this information.
 ;;;
 ;;; These commands are similar to the c-sh-A/m-. commands from the MIT Lisp
 ;;; Machine only much much more sophisticated. The current interface is
-;;; extremely simple yet exceedingly powerfull:
+;;; extremely simple yet exceedingly powerful:
 ;;;
 ;;; m-sh-V  If the cursor points to a variable reference or assignment (SET!)
 ;;;         expression, move the cursor to the defining occurrence for that
@@ -97,7 +97,7 @@
 ;;; high-level Scheme debugger on top of a low-level C debugger like gdb/gud.
 ;;; I don't think that it would be difficult to:
 ;;;
-;;; - allow source level Scheme tracing and single stepping during execution by
+;;; - allow source-level Scheme tracing and single stepping during execution by
 ;;;   moving the coursor around the Scheme program
 ;;;
 ;;; - allow breakpoints
@@ -226,9 +226,10 @@
 (defun stalin:displaced-vector-type-displaced-vector-type (u)
  (nth 21 (stalin:lookup-type u)))
 (defun stalin:type-use-count (u) (nth 23 (stalin:lookup-type u)))
-(defun stalin:has-valuep (u) (eq (nth 24 (stalin:lookup-type u)) 'T))
+(defun stalin:type-fictitiousp (u) (eq (nth 24 (stalin:lookup-type u)) 'T))
 (defun stalin:type-set-members (w) (nth 0 (stalin:lookup-type-set w)))
-(defun stalin:fictitiousp (w) (eq (nth 2 (stalin:lookup-type-set w)) 'T))
+(defun stalin:type-set-fictitiousp (w)
+ (eq (nth 2 (stalin:lookup-type-set w)) 'T))
 (defun stalin:voidp (w) (eq (nth 3 (stalin:lookup-type-set w)) 'T))
 (defun stalin:monomorphicp (w) (eq (nth 4 (stalin:lookup-type-set w)) 'T))
 (defun stalin:multimorphicp (w) (eq (nth 5 (stalin:lookup-type-set w)) 'T))
@@ -244,15 +245,15 @@
  (nth 3 (stalin:lookup-variable g)))
 (defun stalin:variable-name (g) (nth 5 (stalin:lookup-variable g)))
 (defun stalin:variable-environment (g) (nth 6 (stalin:lookup-variable g)))
-(defun stalin:variable-accessedp (g)
- (eq (nth 7 (stalin:lookup-variable g)) 'T))
+(defun stalin:accessedp (g) (eq (nth 7 (stalin:lookup-variable g)) 'T))
 (defun stalin:variable-type-set (g) (nth 8 (stalin:lookup-variable g)))
-(defun stalin:variable-references (g) (nth 9 (stalin:lookup-variable g)))
-(defun stalin:variable-accesses (g) (nth 10 (stalin:lookup-variable g)))
-(defun stalin:globalp (g) (eq (nth 11 (stalin:lookup-variable g)) 'T))
+(defun stalin:accesses (g) (nth 9 (stalin:lookup-variable g)))
+(defun stalin:assignments (g) (nth 10 (stalin:lookup-variable g)))
+(defun stalin:references (g) (nth 11 (stalin:lookup-variable g)))
 (defun stalin:localp (g) (eq (nth 12 (stalin:lookup-variable g)) 'T))
-(defun stalin:slottedp (g) (eq (nth 13 (stalin:lookup-variable g)) 'T))
+(defun stalin:globalp (g) (eq (nth 13 (stalin:lookup-variable g)) 'T))
 (defun stalin:hiddenp (g) (eq (nth 14 (stalin:lookup-variable g)) 'T))
+(defun stalin:slottedp (g) (eq (nth 15 (stalin:lookup-variable g)) 'T))
 (defun stalin:environment-expression (e) (nth 1 (stalin:lookup-environment e)))
 (defun stalin:environment-name (e) (nth 2 (stalin:lookup-environment e)))
 (defun stalin:environment-has-regionp (e)
@@ -261,15 +262,15 @@
 (defun stalin:environment-allocation (e) (nth 5 (stalin:lookup-environment e)))
 (defun stalin:environment-reentrantp (e)
  (eq (nth 6 (stalin:lookup-environment e)) 'T))
-(defun stalin:environment-called-only-oncep (e)
+(defun stalin:environment-called-more-than-oncep (e)
  (eq (nth 7 (stalin:lookup-environment e)) 'T))
 (defun stalin:environment-has-environment-levelp (e)
  (eq (nth 8 (stalin:lookup-environment e)) 'T))
 (defun stalin:environment-unique-call-site (e)
  (nth 9 (stalin:lookup-environment e)))
 
-(defun stalin:findable-variable-references (g)
- (let ((xs1 (stalin:variable-references g))
+(defun stalin:findable-references (g)
+ (let ((xs1 (stalin:references g))
        (xs2 '()))
   (while xs1
    (cond
@@ -291,8 +292,8 @@
   (while (not (= (car is) i)) (setq is (cdr is)) (setq n (+ n 1)))
   n))
 
-(defun stalin:findable-variable-drivers (g)
- (let* ((xs1 (stalin:variable-references g))
+(defun stalin:findable-drivers (g)
+ (let* ((xs1 (stalin:references g))
 	(xs2 '())
 	(e (stalin:variable-environment g))
 	;; needs work
@@ -352,15 +353,16 @@
  (cond
   ((eq major-mode 'scheme-mode)
    (let ((pathname
-	  (concat (stalin:remove-extension (buffer-file-name)) ".db"))
-	 (buffer (current-buffer)))
+          (concat (stalin:remove-extension (buffer-file-name)) ".db"))
+         (buffer (current-buffer)))
     (cond
      ((file-exists-p pathname)
-      (find-file pathname)
-      (setq stalin:*db* (read (current-buffer)))
-      (kill-buffer (current-buffer))
-      (stalin:munge-expressions (nth 0 stalin:*db*) buffer)
-      (stalin:munge-variables (nth 3 stalin:*db*) buffer)))))))
+      (save-excursion
+       (find-file pathname)
+       (setq stalin:*db* (read (current-buffer)))
+       (kill-buffer (current-buffer))
+       (stalin:munge-expressions (nth 0 stalin:*db*) buffer)
+       (stalin:munge-variables (nth 3 stalin:*db*) buffer))))))))
 
 (defun stalin:find ()
  (catch 'return
@@ -564,15 +566,8 @@
  (stalin:externalize-type-set-internal w '()))
 
 (defun stalin:show-type-set (w)
- (let* ((type-set (format "%S" (stalin:externalize-type-set w)))
-	(buffer (current-buffer))
-	(stalin (get-buffer-create "*stalin*")))
-  (call-process "pp" nil stalin nil type-set)
-  (set-buffer stalin)
-  (let ((string (buffer-string)))
-   (set-buffer buffer)
-   (kill-buffer stalin)
-   (lisp-display-output string))))
+ (lisp-display-output
+  (with-output-to-string (pp (stalin:externalize-type-set w)))))
 
 (defun stalin:goto-variable (g)
  (let ((p (stalin:variable-character-position g)))
@@ -621,7 +616,7 @@
  ;;               structure-type-immediate
  ;;               allocating-expressions
  ;;               type-use-count
- (concat (if (stalin:has-valuep u) "has value, " "")
+ (concat (if (stalin:type-fictitiousp u) "fictitious, " "")
 	 "u" (number-to-string u)))
 
 (defun stalin:type-set-info (w)
@@ -629,7 +624,7 @@
  ;;               type-set-has-unionp
  ;;               type-set-multimorphic
  (concat (cond ((stalin:voidp w) "void, ")
-	       ((stalin:fictitiousp w) "fictitious, ")
+	       ((stalin:type-set-fictitiousp w) "fictitious, ")
 	       ((stalin:monomorphicp w) "monomorphic, ")
 	       ((stalin:tag-onlyp w) "tag only, ")
 	       ((stalin:squeezedp w) "squeezed, ")
@@ -638,12 +633,23 @@
 	 "w" (number-to-string w)))
 
 (defun stalin:variable-info (g)
- (concat (if (stalin:variable-accessedp g) "accessed, " "")
-	 (if (stalin:globalp g) "global, " "")
+ (concat (if (stalin:accessedp g) "accessed, " "")
 	 (if (stalin:localp g) "local, " "")
-	 (if (stalin:slottedp g) "slotted, " "")
+	 (if (stalin:globalp g) "global, " "")
 	 (if (stalin:hiddenp g) "hidden, " "")
-	 "a" (number-to-string g) ", "
+	 (if (stalin:slottedp g) "slotted, " "")
+	 (cond ((stalin:localp g) "a")
+	       ((stalin:globalp g) "a")
+	       ((stalin:slottedp g) "a")
+	       (t ""))
+	 (cond ((stalin:localp g) (number-to-string g))
+	       ((stalin:globalp g) (number-to-string g))
+	       ((stalin:slottedp g) (number-to-string g))
+	       (t ""))
+	 (cond ((stalin:localp g) ", ")
+	       ((stalin:globalp g) ", ")
+	       ((stalin:slottedp g) ", ")
+	       (t ""))
 	 "w" (number-to-string (stalin:variable-type-set g))))
 
 (defun stalin:environment-info (e)
@@ -676,7 +682,7 @@
 	   (error "Can't find variable"))
        (stalin:goto-variable (stalin:expression-variable x)))
       (let* ((g (car (cdr x/g)))
-	     (xs (stalin:findable-variable-references g)))
+	     (xs (stalin:findable-references g)))
        (if (null xs) (error "No findable references"))
        (stalin:goto-expression (car xs))))))
 
@@ -691,12 +697,11 @@
 	   (error "Point not on access or assignment"))
        (let ((x (stalin:next
 		 x
-		 (stalin:findable-variable-references
-		  (stalin:expression-variable x)))))
+		 (stalin:findable-references (stalin:expression-variable x)))))
 	(if (not x) (error "Last reference"))
 	(stalin:goto-expression x)))
       (let* ((g (car (cdr x/g)))
-	     (xs (stalin:findable-variable-references g)))
+	     (xs (stalin:findable-references g)))
        (if (null xs) (error "No findable references"))
        (stalin:goto-expression (car xs))))))
 
@@ -711,12 +716,11 @@
 	   (error "Point not on access or assignment"))
        (let ((x (stalin:previous
 		 x
-		 (stalin:findable-variable-references
-		  (stalin:expression-variable x)))))
+		 (stalin:findable-references (stalin:expression-variable x)))))
 	(if (not x) (error "First reference"))
 	(stalin:goto-expression x)))
       (let* ((g (car (cdr x/g)))
-	     (xs (stalin:findable-variable-references g)))
+	     (xs (stalin:findable-references g)))
        (if (null xs) (error "No findable references"))
        (stalin:goto-expression (car xs))))))
 
@@ -800,7 +804,7 @@
        (stalin:goto-variable (stalin:expression-variable x))
        (stalin:goto-first-driver))
       (let* ((g (car (cdr x/g)))
-	     (xs (stalin:findable-variable-drivers g)))
+	     (xs (stalin:findable-drivers g)))
        (if (null xs) (error "No findable drivers"))
        (setq stalin:*xs* xs)
        (let ((x (car xs)))
